import java.awt.*;

class Signal
  implements DeblockListener
{
  static private final boolean DEBUG = false;
  int idx;

  GleisStueck gleis;

  static final int FREE = 1, BLOCKED = 2, EXCEPTION = 3, RESOLVED = 4;
  int status;

  DeblockListener deblockListener;

  Signal(int idx, GleisStueck gleis, DeblockListener deblockListener)
  {
    this.idx = idx;
    this.deblockListener = deblockListener;
    this.gleis = gleis;
    status = FREE;
  }

  private String status2str()
  {
    switch(status)
      {
        case FREE: return "FREE";
	case BLOCKED: return "BLOCKED";
	case RESOLVED: return "RESOLVED";
	default: return "EXCEPTION";
      }
  }

  public synchronized void deblock()
  {
    if(DEBUG) System.out.print("Signal" + idx + "(" +
                               status2str() + ") deblocked -> ");

    if(status == EXCEPTION)
      {
	if(Resources.instance().version == 1)
	  {
	    if(DEBUG) System.out.println("RESOLVED");
            status = RESOLVED;
	  }
	else
	  {
	    if(DEBUG) System.out.println("BLOCKED");
            deblockListener.deblock();
            status = BLOCKED;
	  }
      }
    else
    if(status == RESOLVED)		// only in version = 1
      {
	if(DEBUG) System.out.println("FREE");
	deblockListener.deblock();
        status = FREE;
      }
    else
      {
	if(DEBUG) System.out.println("FREE");
        status = FREE;
      }
  }

  void paint(Graphics g)
  {
    int x = gleis.start.x, y = gleis.start.y, y0;
    // if(gleis.start.alpha==0)
      { y -= 80; y0 = y+20; } 
    // else { y +=60; y0 = y-50; } 

    g.setColor(Color.black);
    g.fillRect(x,y,20,20);
    g.fillRect(x+7,y,6,55);
  }

  void paintLampe(Graphics g)
  {
    int x = gleis.start.x, y = gleis.start.y, y0;
    // if(gleis.start.alpha==0)
      { y -= 80; y0 = y+20; } 
    // else { y +=60; y0 = y-50; } 

    if(status == FREE) g.setColor(Color.green); else g.setColor(Color.red);
    g.fillOval(x+3,y+3,14,14);
  }

  static final int stopdist=1;

  void notifyLokMoved(Lok lok, int oldidx, GleisIterator it)
  {
    int newidx = it.index();
    if((gleis.index-stopdist>oldidx) &&
       (newidx >= gleis.index - stopdist))
      {
	if(status == FREE)
	  {
            if(DEBUG) System.out.println("Signal" + idx + "(FREE) reached.");
	    lok.controller = new PassGreenSignalController(lok, this, it);
	  }
	else
	if(status == BLOCKED || status == RESOLVED)
	  {
            if(DEBUG)
	      System.out.println("Signal" + idx + "(BLOCKED) reached.");
            lok.controller = new StopAtRedSignalController(lok, this, it);
	  }
	else
	  {
            if(DEBUG)
	      System.out.println("Signal" + idx + "(EXCEPTION) reached.");
	    lok.controller = new StopAtExceptionController(lok, this, it);
	  }
      }
    else
    if(gleis.index>oldidx && newidx >= gleis.index)
      {
	if(status == FREE)
	  {
	    status = BLOCKED;
            if(DEBUG)
	      System.out.println(
	        "Signal" + idx + "(FREE) passed -> BLOCKED");

	    deblockListener.deblock();
	  }
	else
        if(status == BLOCKED || status == RESOLVED)
	  {
	    status = EXCEPTION;
            if(DEBUG)
	      System.out.println("Signal" + idx + "(BLOCKED) passed -> " +
	                         status2str());
	    lok.controller = new PassRedSignalController(lok, this, it);
	  }
      }
  }
}
