import java.awt.*;

class ZugStrecke extends Canvas
  implements Runnable
{
  private static final boolean DEBUG = false;

  private LokManager lokmanager = new LokManager();

  private GleisStueck gleis;
  private Dimension _preferredSize;
  private Signal[] signal = new Signal[6];
  private Signal lastsig;

  private Image background, tunnel, bang;
  private int width, height, mouse_x=-1000, mouse_y=-1000,
    mouse_down_x, mouse_down_y, tunnel_y, tunnel_x;

  boolean running=false;
  Turtle collision_pos = null;
  int train_moved_over_red_signal = -1;
  private int velocity=2;

  private static int number_of_threads=0;
  private ResetListener reset_listener;
  private DeblockListener deblock_listener;

  private int endidx;

  ZugStrecke(ResetListener rl, DeblockListener dl)
  {
     reset_listener = rl;
     deblock_listener = dl;

     Resources.instance().loadImages();

     background=Resources.instance().getImage("green.gif");
     tunnel=Resources.instance().getImage("gray.gif");
     bang=Resources.instance().getImage("bang.gif");

     Turtle start = new Turtle(0, 100, 0);

     gleis = new GeradeRight(start);
     GleisStueck s = new GeradeRight(gleis);
     s = new GeradeRight(s);
     s = new GeradeRight(s);
     int signum=0;
     signal[signum] = new Signal(signum,s, dl);
     s = new GeradeRight(s);
     s = new GeradeRight(s);
     s = new GeradeRight(s);
     s = new KurveNE(s);
     s = new GeradeDown(s);
     s = new KurveES(s);
     s = new GeradeLeft(s);
     s = new GeradeLeft(s);
     s = new GeradeLeft(s);
     signum++; signal[signum] = new Signal(signum,s, signal[signum-1]);
     s = new GeradeLeft(s);
     s = new GeradeLeft(s);
     s = new KurveNW(s);
     s = new GeradeDown(s);
     s = new GeradeDown(s);
     s = new KurveWS(s);
     s = new GeradeRight(s);
     s = new GeradeRight(s);
     s = new GeradeRight(s);
     s = new GeradeRight(s);

     if(Resources.instance().large)
       {
         s = new GeradeRight(s);
         signum++; signal[signum] = new Signal(signum,s, signal[signum-1]);
         s = new GeradeRight(s);
         s = new GeradeRight(s);
         s = new GeradeRight(s);
	 s = new KurveSE(s);
	 s = new GeradeUp(s);
	 s = new GeradeUp(s);
	 s = new GeradeUp(s);
	 s = new GeradeUp(s);
	 s = new GeradeUp(s);
	 s = new KurveWN(s);
         s = new GeradeRight(s);
         s = new KurveNE(s);
         signum++; signal[signum] = new Signal(signum,s, signal[signum-1]);
	 s = new GeradeDown(s);
	 s = new GeradeDown(s);
	 s = new GeradeDown(s);
	 s = new GeradeDown(s);
	 s = new GeradeDown(s);
	 s = new KurveWS(s);
         s = new GeradeRight(s);
         s = new GeradeRight(s);
       }

     signum++;
     lastsig = signal[signum] = new Signal(signum,s, signal[signum-1]);

     s = new GeradeRight(s);
     endidx = s.index;
     tunnel_x = s.end.x; tunnel_y = s.end.y;
     s = new GeradeRight(s);

     _preferredSize = new Dimension(gleis.max_x()+100, gleis.max_y()+100);
  }

  synchronized void newLok()
  {
    Lok l = lokmanager.newLok(gleis.start);
    l.controller = new UserController(l);
    (new Thread(this)).start();
  }

  public Dimension preferredSize() { return _preferredSize; }
  public Dimension minimumSize() { return preferredSize(); }

  private synchronized void _drawSelections(Graphics g)
  {
    if(g==null || lokmanager.lok==null) return;
    for(int i=0; i<lokmanager.lok.length; i++)
      lokmanager.lok[i].drawSelection(g);
  }

  private synchronized void drawSelections()
  {
    Graphics g = this.getGraphics();
    _drawSelections(g);
  }

  private synchronized void setSelection()
  {
    if(lokmanager.lok==null) return;

    if(mouse_x>=0 && mouse_y>=0)
      {
        for(int i=0; i<lokmanager.lok.length; i++)
	  if(lokmanager.lok[i].boundingBox().inside(mouse_x, mouse_y))
            lokmanager.lok[i].controller.select();
	  else lokmanager.lok[i].controller.deselect();
      }
    else
      {
        for(int i=0; i<lokmanager.lok.length; i++)
          lokmanager.lok[i].controller.deselect();
      }

    // drawSelections();
  }

  private synchronized void setPressed()
  {
    if(lokmanager.lok==null) return;

    if(mouse_x>=0 && mouse_y>=0)
      {
        for(int i=0; i<lokmanager.lok.length; i++)
	  if(lokmanager.lok[i].boundingBox().inside(mouse_x, mouse_y))
            lokmanager.lok[i].controller.press();
	  else lokmanager.lok[i].controller.release();
      }

    // drawSelections();
  }
  
  private synchronized void releasePressed()
  {
    if(lokmanager.lok==null) return;

    for(int i=0; i<lokmanager.lok.length; i++)
       lokmanager.lok[i].controller.release();

    // drawSelections();
  }

  private synchronized int numberOfMouseOnLok()
  {
    if(lokmanager.lok==null) return -1;

    if(mouse_x>=0 && mouse_y>=0)
      {
        for(int i=0; i<lokmanager.lok.length; i++)
	  if(lokmanager.lok[i].boundingBox().inside(mouse_x, mouse_y))
	    return i;
      }

    return -1;
  }

  /*
  int getDirection(Turtle t)
  {
     double angle =
       Math.atan2(mouse_y - mouse_down_y, mouse_x - mouse_down_x);
     angle *= 180;
     angle /= Math.PI;
     if(angle<0) angle += 360;
     Assert.assert(0 <= angle && angle <=360);
     if(Math.abs(angle - t.alpha) > 90) return -1;
     else return 1;
  }
  */

  public boolean handleEvent(Event e)
  {
    switch(e.id)
      {
	case Event.MOUSE_ENTER:
	  requestFocus();
	  // FALL THROUGH

	case Event.MOUSE_MOVE:
	  mouse_x = e.x;
	  mouse_y = e.y;
	  setSelection();
	  return true;
	
	case Event.MOUSE_EXIT:
	  mouse_x = mouse_y = -1000;
	  setSelection();
	  return true;

        case Event.MOUSE_DOWN:
	  {
	    mouse_down_x = mouse_x = e.x;
	    mouse_down_y = mouse_y = e.y;
	    int n = numberOfMouseOnLok();
	    if(n>=0)
	      {
		if(lokmanager.lok[n].isRunning())
		  lokmanager.lok[n].controller.stop();
		else lokmanager.lok[n].controller.go();
		setPressed();
		return true;
	      }
	  }
	  break;
	
	/*
	case Event.MOUSE_DRAG:
	  {
	    mouse_x = e.x;
	    mouse_y = e.y;
	    int n = numberOfMouseOnLok();
	    if(n>=0)
	      {
	        if(!lokmanager.lok[n].running)
		  {
		    int direction = getDirection(lokmanager.lok[n].pos());
		    if(direction<0)
		      {
			if(lokmanager.lok[n].velocity>1)
			  lokmanager.lok[n].velocity--;
			else lokmanager.lok[n].direction *= -1;
		      }
		    else lokmanager.lok[n].velocity++;

		    lokmanager.lok[n].running = true;
		  }
		return true;
	      }
	  }
	  break;
	*/
	
	case Event.MOUSE_UP:
	  mouse_x = e.x;
	  mouse_y = e.y;
	  releasePressed();
	  return true;
	
	case Event.KEY_PRESS:
	  {
	    /*
	    if(DEBUG) System.out.println(
	      "Zugstrecke.handleEvent(KEY_PRESS," + e.key + ")");
	    */

	    int n = numberOfMouseOnLok();
	    if(n>=0)
	      {
		Resources rsrc = Resources.instance();
		if(e.key == rsrc.faster_char)
		  {
		    lokmanager.lok[n].controller.incVelocity();
		    return true;
		  }
		else
		if(e.key == rsrc.slower_char)
		  {
		    lokmanager.lok[n].controller.decVelocity();
		    return true;
		  }
		else
		if(e.key == rsrc.stop_char)
		  {
		    lokmanager.lok[n].controller.stop();
		    return true;
		  }
		else
		if(e.key == rsrc.go_char)
		  {
		    lokmanager.lok[n].controller.go();
		    return true;
		  }
		else
		if(e.key == ' ')
		  {
		    lokmanager.lok[n].controller.toggleDirection();
		    return true;
		  }
	      }
	  }
          break;

        default:
          break;
      }

    return false;
  }

  private static synchronized int myThreadNumber()
  {
    return number_of_threads++;
  }

  public void run()
  {
    Image offscreen = null;	// for each an own offscreen image!
    int my_thread_number = myThreadNumber();

    Lok lok=lokmanager.lastLok();

    while(getGraphics() == null)
      { try {Thread.sleep(1000);} catch(InterruptedException e) {}; }

    if(DEBUG)
      System.out.println("Animation" + my_thread_number + " started.");
    

    GleisIterator it = new GleisIterator(gleis);

    it.first();

    while(!lok.remove &&
          (collision_pos == null) &&
          (train_moved_over_red_signal<0) &&
	  !it.isDone())
      {
        /*
        if(DEBUG) System.out.println(
          "Animation" + my_thread_number + " at " + it.turtle());
        */

        Dimension d = size();
        while(d.height<=0 || d.width<=0)
          try {Thread.sleep(1000);} catch(InterruptedException e) {};

        if((offscreen == null) ||
           (width != d.width) ||
           (height != d.height))
          {
            /*
            offscreen = createImage(d.width, d.height);
            */

            while((offscreen = createImage(d.width, d.height))==null)
              try {Thread.sleep(1000);} 
              catch(InterruptedException e) {};

            /*
            Assert.assert(offscreen != null,
              "could not create offscreen image for Animation" +
              my_loknumber);
            */

            width = d.width;
            height = d.height;

            if(DEBUG && offscreen!=null) System.out.println(
              "Animation" + my_thread_number +
              " offscreen(" + width + "," + 
              height + ") image created");
          }
         

        setSelection();
        if(offscreen==null)
          {
            Graphics g = this.getGraphics();
            paint(g);
          }
        else
          {
            Rectangle oldrect = lok.boundingBox();
            lok.move(it.turtle());
            Rectangle newrect =  lok.boundingBox();
            Rectangle r = newrect.union(oldrect);

            Graphics g = offscreen.getGraphics();
            g.clipRect(r.x, r.y, r.width, r.height);
            paint(g);
            g = this.getGraphics();
            g.clipRect(r.x, r.y, r.width, r.height);
            g.drawImage(offscreen, 0, 0, this);
          }

        try { Thread.sleep(50); }
        catch(InterruptedException e) { }

	int oldidx = it.index();
        if(running) it.next(lok.controller.request(velocity));

	if(oldidx<endidx && it.index() == endidx)
	  {
	    lastsig.deblock();
	    lok.controller = new RunTillTheEndController(lok);
	  }

	for(int i=0; i<signal.length; i++)
	  if(signal[i]!=null) signal[i].notifyLokMoved(lok, oldidx, it);

	if(collision_pos == null)
	  collision_pos = lokmanager.collision();
	
	if(collision_pos != null)
	  {
	    if(DEBUG) System.out.println("Zugstrecke Kollision!");
	    reset_listener.requestReset();
	  }
	else
	if(Resources.instance().version == 3)	// uncommented !!
	  {
	    for(int i=0;
	        train_moved_over_red_signal<0 && i<signal.length;
		i++)
	      {
	        if(signal[i] != null && signal[i].status == Signal.EXCEPTION)
	          train_moved_over_red_signal = i;
	      }
	    
	    if(train_moved_over_red_signal>=0)
	      {
	        if(DEBUG) System.out.println(
		  "Signal" + train_moved_over_red_signal + " Ueberfahren!");

		reset_listener.requestReset();
	      }
	  }
      }
    
    if(collision_pos != null)
      {
        paintBang();
	lokmanager.removeLok(lok);
      }
    else
    if(train_moved_over_red_signal>=0)
      {
        paintTrainMovedOverRedSignal();
	lokmanager.removeLok(lok);
      }
    else
      {
        Rectangle r = lok.boundingBox();
        lokmanager.removeLok(lok);
        Graphics g = offscreen.getGraphics();
        g.clipRect(r.x, r.y, r.width, r.height);
        paint(g);
        g = this.getGraphics();
        g.clipRect(r.x, r.y, r.width, r.height);
        g.drawImage(offscreen, 0, 0, this);
      }
    
    if(DEBUG)
      System.out.println("Animation" + my_thread_number + " finished.");
  }

  private void paintBang()
  {
    if(collision_pos == null) return;
    Graphics g = this.getGraphics();
    int x = collision_pos.x - 58, y = collision_pos.y - 36, w = 117, h = 72;
    g.clipRect(x,y,w,h);
    g.drawImage(bang, x, y, this);
  }

  private void paintTrainMovedOverRedSignal()
  {
    int i = train_moved_over_red_signal;
    if(i<0) return;
    Graphics g = this.getGraphics();
    g.setFont(new Font("Helvetica", Font.BOLD, 40));
    g.setColor(Color.red);

    FontMetrics fm = getFontMetrics(g.getFont());
    int d = fm.getHeight() + fm.getDescent();
    int x = signal[i].gleis.start.x - 10;

    String msg = "Signal";
    int y = signal[i].gleis.start.y - 60 + d;
    g.drawString(msg, x - fm.stringWidth(msg)/2, y);

    msg = "Ueberfahren!";
    g.drawString(msg, x - fm.stringWidth(msg)/2, y + d);
  }

  void reset()
  {
    velocity = 2;
    lokmanager.removeAllLoks();
    collision_pos = null;
    train_moved_over_red_signal = -1;
    deblock_listener.deblock();

    for(int i=0; i<signal.length; i++)
      if(signal[i]!=null) signal[i].status = Signal.FREE;

    repaint();
  }

  void faster() { velocity++; } 
  void slower() { if(velocity>1) velocity--; }

  synchronized private void drawLoks(Graphics g)
  {
    if(lokmanager.lok != null)
      for(int i=0; i<lokmanager.lok.length; i++)
        lokmanager.lok[i].paint(g, this);
  }

  private void drawSignale(Graphics g)
  {
    for(int i=0; i<signal.length; i++)
      if(signal[i] != null) signal[i].paint(g);

    g = this.getGraphics();
    for(int i=0; i<signal.length; i++)
      if(signal[i] != null) signal[i].paintLampe(g);
  }

  static private final int COPYRIGHTFONTSIZE = 10;

  public void paint(Graphics g)
  {
    // if(DEBUG) System.out.println("ZugStrecke.paint(...)");

    Dimension sz = size();

    for(int i=0; i<=sz.width; i+=100)
      for(int j=0; j<=sz.height; j+=100)
        g.drawImage(background, i, j, 100, 100, this);

    // Copyright message

    g.setFont(new Font("Helvetica", Font.BOLD, COPYRIGHTFONTSIZE));
    g.setColor(Color.yellow);
    String msg =
      "(C)opyright 1997 Armin Biere, ILKD, University of Karlsruhe";
    FontMetrics fm = getFontMetrics(g.getFont());
    g.drawString(msg, sz.width - fm.stringWidth(msg) - COPYRIGHTFONTSIZE,
		      sz.height - fm.getAscent() - COPYRIGHTFONTSIZE);
    msg =
      "(C)opyright 2001 Armin Biere, Computer Systems Institute, ETH Zurich";
    g.drawString(msg, sz.width - fm.stringWidth(msg) - COPYRIGHTFONTSIZE,
		      sz.height - fm.getAscent());
    
    // Version message

    g.setColor(Color.white);
    fm = getFontMetrics(g.getFont());
    msg = " Version 1." + Resources.instance().version;
    g.drawString(msg, 0, sz.height - fm.getAscent());
			       
    gleis.paint(g, this);
    drawSignale(g);
    drawLoks(g);

    g.drawImage(tunnel, tunnel_x, tunnel_y - 50, this);
    g.drawImage(tunnel, tunnel_x + 100, tunnel_y - 50, this);
    g.drawImage(tunnel, tunnel_x + 200, tunnel_y - 50, this);

    // Selection

    _drawSelections(g);
    paintBang();
    paintTrainMovedOverRedSignal();
  }
}
